/*
 * libgsmat: An implementation of GSM 07.07
 *
 * Written by mark.liu <mark.liu@openvox.cn>
 *
 * Copyright (C) 2009-2010 OpenVox
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include "libgsmat.h"
#include "gsm_internal.h"
#include "m20.h"

static int m20_trim(char *in, char *out, int len) {
    int i=0;
    int j=0;

    for (i = 0; i < len; i++) {
        if ((in[i] != '\r') && (in[i] != '\n') ) {
            out[j++] = in[i];
        }
    }
    out[j++] = 0x0;
    return j;
}

static int m20_transmit(struct gsm_modul *gsm, char *data, int len) {
    int res;

    if (gsm->debug) {
        gsm_message(gsm, "> %s\n", data);
    }
    m20_trim(data, gsm->lastcmd, len);
    res = write(gsm->fd, data, len+2);
    if (res != (len+2)) {
        gsm_error(gsm, "Short write: %d/%d (%s) on span %d\n", res, len+2, strerror(errno), gsm->span);
        return -1;
    }
    return 0;
}

static int m20_send(struct gsm_modul *gsm, char *data) {
    int len = 0;

    if (!data) {
        return -1;
    }

    len = strlen(data);

    return m20_transmit(gsm, data, len);
}

void m20_send_pin(struct gsm_modul *gsm, char *pin) {
    char buf[256];
    memset(buf, 0x0, sizeof(buf));
    // If the SIM PIN is blocked
    if (gsm->state == GSM_STATE_SIM_PIN_REQ) {
        snprintf(buf, sizeof(buf) - 1, "AT+CPIN=\"%s\"\r", pin); //eg: AT+CPIN="1234"
        m20_send(gsm, buf);
    }
}

static int m20_switch_state(struct gsm_modul *gsm, int state, char *next_command)
{
    gsm->state = state;
    if (next_command) {
        m20_send(gsm, next_command);
    }
    return 0;
}

static int m20_switch_sim_state(struct gsm_modul *gsm, int state, char *next_command)
{
    gsm->sim_state = state;
    if (next_command) {
        m20_send(gsm, next_command);
    }

    return 0;
}
static void m20_error_hard(void *data) {
    struct gsm_modul *gsm = (struct gsm_modul *)data;
    gsm->restart_timer = 0;
    gsm->state = GSM_STATE_ERROR;
    gsm->ev.e = GSM_EVENT_ERROR;
    gsm->ev.error.hard = 1;
    gsm->ev.error.cause = 44; /* AST_CAUSE_REQUESTED_CHAN_UNAVAIL */
    gsm->schedev = 1;
}

int m20_request_status(struct gsm_modul *gsm) {
    char buf[256];
    if (gsm->state <= GSM_STATE_READY) {
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CSQ\r");
        gsm->state = GSM_STATE_STATUS_REQ;
        m20_transmit(gsm, buf, strlen(buf));
        return 0;
    } else {
        gsm_error(gsm, "cannot request GSM status, module is in use\n");
        return -1;
    }

    return -1;
}

void m20_start(struct gsm_modul *gsm) {
    char buf[256];
    gsm->state = GSM_STATE_POWER_ON;
    memset(buf, 0x0, sizeof(buf));
    snprintf(buf, sizeof(buf) - 1, "ATZ\r");
    m20_transmit(gsm, buf, strlen(buf));
}

void m20_restart(void *data) {
    struct gsm_modul *gsm = (struct gsm_modul *)data;
    if (gsm->debug) {
        gsm_message(gsm, "restarting\n");
    }
    gsm->dchanup = 0;
    gsm->state = GSM_STATE_POWER_ON;
    gsm->restart_timer = gsm_schedule_event(gsm, 10000, m20_error_hard, gsm);
    m20_start(gsm);
}

int m20_answer(struct gsm_modul *gsm) {
    if (gsm->state == GSM_STATE_RINGING) {
        m20_send(gsm, "ATA\r");
    }
    return 0;
}

int m20_dial(struct gsm_modul *gsm, int clir, char *called) {
    char buf[256];
    if (gsm->state == GSM_STATE_READY) {
        gsm->hungup = 0;
        if (clir > 0) {
            clir = 1;
        }
        gsm->clir = clir;
        strncpy(gsm->cpn, called, sizeof(gsm->cpn) - 1);
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CLIR=%d\r", clir);
        gsm->state = GSM_STATE_CALL_INIT;
        m20_send(gsm, buf);
        return 0;
    }
    gsm_error(gsm, "Cannot dial when in state %d\n", gsm->state);
    return -1;
}

int m20_sms_send_text(struct gsm_modul *gsm, char *destination, char *msg) {
    char buf[1024];
    if (gsm->state == GSM_STATE_READY) {
        snprintf(gsm->msgbuf, sizeof(gsm->msgbuf) - 1, "%s", msg);

        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CMGF=1\r");
        m20_transmit(gsm, buf, strlen(buf));
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CMGS=\"%s\"\r", destination);
        m20_switch_state(gsm, GSM_STATE_SM_SEND_REQ, buf);
        return 0;
    }
    gsm_error(gsm, "Cannot send SM when in state %d\n", gsm->state);
    return -1;
}

int m20_sms_send_pdu(struct gsm_modul *gsm, char *pdu) {
    char buf[1024];
    char smsc[3];
    int smsc_len;
    int len = 0;

    if (!pdu) return 0;

    len = strlen(pdu);

    /* SMSC information length */
    strncpy(smsc, pdu, sizeof(smsc) - 1);
    smsc[2] = '\0';
    smsc_len = gsm_hex2int(smsc, 2);
    len = (len / 2) - 1 - smsc_len;

    if (gsm->state == GSM_STATE_READY) {
        snprintf(gsm->msgbuf, sizeof(gsm->msgbuf) - 1, "%s", pdu);

        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CMGF=0\r");
        m20_transmit(gsm, buf, strlen(buf));
        memset(buf, 0x0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AT+CMGS=%d\r", len);
        m20_transmit(gsm, buf, strlen(buf));

        gsm->state = GSM_STATE_SM_SEND_REQ;
        return 0;
    }
    gsm_error(gsm, "Cannot send SM when in state %d\n", gsm->state);
    return -1;
}


int m20_hangup(struct gsm_modul *gsm) {
    char buf[256];
    gsm->cpn[0] = '\0';
    memset(buf, 0x0, sizeof(buf));
    snprintf(buf, sizeof(buf) - 1, "ATH\r");
    if (!gsm->hungup && ((gsm->state == GSM_STATE_CALL_PRESENT) || (gsm->state == GSM_STATE_CALL_ACTIVE) || (gsm->state == GSM_STATE_RINGING))) {
        gsm->state = GSM_STATE_HANGUP_REQ;
        m20_transmit(gsm, buf, strlen(buf));
        if (!gsm->restart_timer) {
            gsm->restart_timer = gsm_schedule_event(gsm, 3000, m20_restart, gsm);
        }
    }
    return 0;
}

static int m20_san(struct gsm_modul *gsm, char *in, char *out, int len) {
    int i=0;
    int skip=0;
    char *tmp = NULL;

    if ((len > 0) && ((gsm->sanidx + len) < sizeof(gsm->sanbuf))) {
        memcpy(gsm->sanbuf + gsm->sanidx, in, len);
        gsm->sanidx += len;
        gsm->sanbuf[gsm->sanidx] = 0x0;
    }

    while ((gsm->sanbuf[i] == '\r') || (gsm->sanbuf[i] == '\n')) {
        i++;
    }

    skip = i;

    tmp = strchr(gsm->sanbuf + skip, '\r');

    if (tmp){
        i = tmp - (gsm->sanbuf + skip);
        memcpy(out, (gsm->sanbuf + skip), i);
        out[i] = 0x0;
        memmove(&gsm->sanbuf[0], &gsm->sanbuf[skip+i], sizeof(gsm->sanbuf) - i - skip);
        gsm->sanidx -= i + skip;
        return i;
    } else {
        if (gsm->state == GSM_STATE_SM_SEND_REQ) {
            if ((gsm->sanidx - skip == 2) && (gsm->sanbuf[gsm->sanidx - 2] == '>') && (gsm->sanbuf[gsm->sanidx - 1] == ' ')) {
                out[0] = '>';
                out[1] = ' ';
                out[2] = 0x0;
                gsm->sanidx -= 2 + skip;
                return 2;
            }
        }
    }

    return 0;
}

gsm_event *m20_receive(struct gsm_modul *gsm, char *data, int len)
{
    int res;
    char buf[1024];
    char buf2[1024];
    char buf3[1024];
    char callerid[256];
    char loc_str[256];
    char *a = NULL;
    int creg1 = 0;
    int creg2 = 0;
    int registered = 0;

    res = m20_san(gsm, data, buf, len);
    if (!res) {
        return NULL;
    }

    m20_trim(buf, buf2, res);

    if (gsm->debug) {
        gsm_message(gsm, "< %s , %d >\n",buf2, res);
    }

    strncpy(buf3, buf2, sizeof(buf3) - 1);

    if (gsm_compare(buf3, "+CREG:")) {
        if (gsm_compare(gsm->lastcmd, "AT+CREG?")) {
           if (sscanf(buf3, "+CREG: %d,%d,%s", &creg1, &creg2, loc_str) == 3) {
            }
            gsm->lastcmd[0] = '\0';
        } else {
            if (gsm_compare(buf3, "+CREG: 002")) {
                /* Not registered, but the ME is currently searching for a
                   new operator to which to register.*/
                creg2 = 2;
            } else if (gsm_compare(buf3, "+CREG: 003")) {
                /* Registration denied. */
                creg2 = 3;
            } else if (sscanf(buf3, "+CREG: %d,%s", &creg1, loc_str) == 2) {
                /* unsolicited */
                creg2 = creg1;
                m20_send(gsm, "AT+CREG=2\r");
            }
        }

        registered = (creg2 == 1) || (creg2 == 5) || (creg2 == 8);
        if (registered) {
           gsm->state = GSM_STATE_READY;
           if (!gsm->dchanup) {
                gsm->dchanup = 1;
                m20_send(gsm, "AT+CLIP=1\r");
                gsm->ev.e = GSM_EVENT_DCHAN_UP;
                return &gsm->ev;
            }
        } else {
            if (gsm->dchanup) {
                gsm->dchanup = 0;
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_DCHAN_DOWN;
                return &gsm->ev;
            }
        }
    }

    switch (gsm->state) {
        case GSM_STATE_POWER_ON:
            if (gsm_compare(buf2, "OK")) {
                if (gsm_compare(gsm->lastcmd, "ATZ")) {
                    sleep(1);
                    m20_switch_state(gsm, GSM_STATE_INIT_0, "ATE0\r");
                }
                if (gsm->restart_timer) {
                    gsm_schedule_del(gsm, gsm->restart_timer);
                    gsm->restart_timer = 0;
                }
            } else if (gsm_compare(buf2, "ATZ")) {
                if (gsm->restart_timer) {
                    gsm_schedule_del(gsm, gsm->restart_timer);
                    gsm->restart_timer = 0;
                }
            } else if (gsm_compare(buf2, "Call Ready")) {
            } else if (gsm_compare(buf2, "ERROR")){
                m20_start(gsm);
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_INIT_0:
            if (gsm_compare(buf2, "OK")) {
                sleep(1); /* To show the network status */
                m20_switch_state(gsm, GSM_STATE_REPORT_ERROR, "AT+CMEE=2\r");
            } else if (gsm_compare(buf2, "ATE0")) {
                if (gsm->restart_timer) {
                    gsm_schedule_del(gsm, gsm->restart_timer);
                    gsm->restart_timer = 0;
                }
            } else if (gsm_compare(buf2, "NO CARRIER") || gsm_compare(buf2, "ERROR")) {
                    gsm->restart_timer = gsm_schedule_event(gsm, 1000, m20_restart, gsm);
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_REPORT_ERROR:
            if (gsm_compare(buf2, "OK")) {
                if (gsm_compare(gsm->lastcmd, "AT+CMEE=2")) {
                    m20_switch_state(gsm, GSM_STATE_VERSION_REQ, "AT+CGMR\r");
                }
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_VERSION_REQ:
            if (gsm_compare(buf2, "OK")) {
                if (gsm_compare(gsm->lastcmd, "AT+CGMR")) {
                    m20_switch_state(gsm, GSM_STATE_IMSI_REQ, "AT+CIMI\r");
                }
            } else if (gsm_compare(buf2, "Revision:")) {
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_IMSI_REQ:
            if (!gsm_compare(buf2, "+CME ERROR:")) {
                if (gsm_compare(gsm->lastcmd, "AT+CIMI")) {
                    gsm_set_imsi(gsm, buf2);
                    if (gsm_compare(buf2, "OK")) {
                        m20_switch_state(gsm, GSM_STATE_IMEI_REQ, "AT+CGSN\r");
                    }
                }
            }
            break;
        case GSM_STATE_IMEI_REQ:
            if (!gsm_compare(buf2, "+CME ERROR:")) {
                if (gsm_compare(gsm->lastcmd, "AT+CGSN")) {
                    gsm_set_imei(gsm, buf2);
                    if (gsm_compare(buf2, "OK")) {
                        m20_switch_state(gsm, GSM_STATE_CHECK_SIM_SECURITY, "AT+CPIN?\r");
                    }
                }
            }
            break;
        case GSM_STATE_CHECK_SIM_SECURITY:
            if (gsm_compare(buf2, "+CPIN: READY")) {
                m20_switch_sim_state(gsm, GSM_STATE_SIM_READY, NULL);
            } else if (gsm_compare(buf2, "+CPIN: SIM PIN")) {
                m20_switch_sim_state(gsm, GSM_STATE_SIM_PIN_REQ, NULL);
            } else if (gsm_compare(buf2, "OK")) {
                switch(gsm->sim_state) {
                case GSM_STATE_SIM_READY:
                    m20_switch_state(gsm, GSM_STATE_SIM_READY, "AT+CREG=2\r");
                    break;
                case GSM_STATE_SIM_PIN_REQ:
                    m20_switch_state(gsm, GSM_STATE_SIM_PIN_REQ, NULL);
                    gsm->ev.e = GSM_EVENT_PIN_REQUIRED;
                    return &gsm->ev;
                    break;
                }
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_SIM_PIN_REQ:
            if (gsm_compare(buf2, "OK")) {
                /* gsm_message(gsm, "LAST>%s<\n",gsm->lastcmd); */
                m20_switch_state(gsm, GSM_STATE_SIM_READY, "AT+CREG=2\r");
            } else if (gsm_compare(buf2, "+CME ERROR: SIM not inserted")) {
                /* re-enter SIM if neccessary */
                gsm->state = GSM_STATE_SIM_PIN_REQ;
                gsm->ev.e = GSM_EVENT_PIN_REQUIRED;
                return &gsm->ev;
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_SIM_READY:
            if (gsm_compare(buf2, "OK")) {
                m20_switch_state(gsm, GSM_STATE_NET_REQ, "AT+CREG?\r");
            } else if (gsm_compare(buf2, "AT+CREG=2")) {
                /* echo */
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_NET_REQ:
            if (gsm_compare(buf2, "+CREG:")) {
            } else if (gsm_compare(buf2, "AT+CREG?")) {
            } else {
            }
            break;
        case GSM_STATE_READY:
            if (gsm_compare(gsm->lastcmd, "AT+CLIP=1")) {
                 // Set number of seconds to wait for connection completion
                usleep(1000);
                m20_send(gsm, "ATS7=255\r");
            } else if (gsm_compare(gsm->lastcmd, "ATS7=255")) {
                usleep(1000);
                m20_send(gsm, "AT+QMOSTAT=1\r"); // show call state of mobile originated call.
            } else if (gsm_compare(gsm->lastcmd, "AT+QMOSTAT")) {
                usleep(1000);
                m20_send(gsm, "AT+QAUDCH=0\r"); // Normal audio channel(default)
            } else if (gsm_compare(gsm->lastcmd, "AT+QAUDCH")) {
                usleep(1000);
                m20_send(gsm, "AT+QMIC=0,2\r");     // Normal Microphone ; 2 (default)
            } else if (gsm_compare(gsm->lastcmd, "AT+QMIC=0,2")) {
                usleep(1000);
                m20_send(gsm, "AT+QSIDET=0\r");     // Change the side tone gain level
            } else if (gsm_compare(gsm->lastcmd, "AT+QSIDET=0")) {
                usleep(1000);
                m20_send(gsm, "AT+QECHO=0,0,0,0\r");// echo chancellation control
            } else if (gsm_compare(gsm->lastcmd, "AT+QECHO")) {
                usleep(2000);
                m20_send(gsm, "AT+CLVL=93\r");
            } else if (gsm_compare(gsm->lastcmd, "AT+CLVL=93")) {
                m20_send(gsm, "AT+CLIR=0\r");
            } else if (gsm_compare(gsm->lastcmd, "AT+CLIR=0")) {
                usleep(1000);
                m20_send(gsm, "AT+CMGF=0\r");
            } else if (gsm_compare(gsm->lastcmd, "AT+CMGF=0")) {
                usleep(1000);
                m20_send(gsm, "AT+CSQ\r");
            } else if (gsm_compare(gsm->lastcmd, "AT+CSQ")) {
                usleep(1000);
                m20_send(gsm, "AT+CNMI=0,2,2,0,0\r");
            }

            callerid[0] = '\0';
            gsm->ev.ring.callingnum[0] = '0';

            if (gsm_compare(buf2, "+CLIP:")) {
                gsm->state = GSM_STATE_RINGING;
                if (sscanf(buf2, "+CLIP: \"%s\"", callerid) == 1) {
                    a = strchr(callerid, '"');
                    if (a) {
                        *a = '\0';
                    }
                }

                gsm->hungup = 0;

                if (!strlen(callerid)) {
                    strncpy(gsm->ev.ring.callingnum, "", sizeof(gsm->ev.ring.callingnum) - 1);
                } else {
                    strncpy(gsm->ev.ring.callingnum, callerid, sizeof(gsm->ev.ring.callingnum) - 1);
                }

                gsm->ev.e = GSM_EVENT_RING;
                return &gsm->ev;
            } else if (gsm_compare(buf2, "+CMT:")) {
                gsm->state = GSM_STATE_SM_RECEIVING;
            } else if (gsm_compare(buf2, "NO CARRIER")) {
                if (gsm_compare(gsm->lastcmd, "ATD")) {
                    gsm->state = GSM_STATE_READY;
                    gsm->ev.e = GSM_EVENT_HANGUP;
                    gsm->ev.hangup.cause = 16;
                    return &gsm->ev;
                }
            } else if (gsm_compare(buf2, "OK")) {
                if (gsm_compare(gsm->lastcmd, "ATD")) {
                    gsm->state = GSM_STATE_CALL_PRESENT;
                } else {
                    /* echo */
                }
            /* sms sent */
            } else if (gsm_compare(buf2, "+CREG")) {
            } else if (gsm_compare(buf2, "RING")) {
            } else if (gsm_compare(buf2, "+CSQ")) {
                /* echo */
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_RINGING:
            if (gsm_compare(buf2, "NO CARRIER")) { // ATA Response if no connection or caller hangup
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            } else if (gsm_compare(buf2, "OK")) {
                if (gsm_compare(gsm->lastcmd, "ATA")) {
                    gsm->state = GSM_STATE_CALL_ACTIVE;
                }
            } else if (gsm_compare(buf2, "BUSY")) {
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 17;
                return &gsm->ev;
            } else if (gsm_compare(buf2, "+CLIP:")) {
            } else if (gsm_compare(buf2, "RING")) {
            } else if (gsm_compare(buf2, "ATA")) {
                /* echo */
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;

        case GSM_STATE_CALL_ACTIVE:
            if (gsm_compare(buf2, "NO CARRIER")) { // callee hangup
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 16;
                return &gsm->ev;
            } else if (gsm_compare(buf2, "OK")) {
                if (gsm_compare(gsm->lastcmd, "ATH")) {
                    gsm->state = GSM_STATE_READY;
                }
            } else if (gsm_compare(buf2, "+CREG")) {
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_CALL_INIT:
            if (gsm_compare(buf2, "AT+CLIR=")) {
            } else if (gsm_compare(buf2, "OK")) {
                if (gsm_compare(gsm->lastcmd, "AT+CLIR=")) {
                    gsm->lastcmd[0] = '\0';
                    if (strlen(gsm->cpn)) {
                        memset(buf, 0x0, sizeof(buf));
                        snprintf(buf, sizeof(buf) - 1, "ATD%s;\r", gsm->cpn); // ATD may be aborted generally by receiving an ATH command.
                        m20_switch_state(gsm, GSM_STATE_CALL_PRESENT, buf);
                    } else {
                        gsm->state = GSM_STATE_READY;
                    }
                }
            }
            break;
        case GSM_STATE_CALL_PRESENT:
            if (gsm_compare(buf2, "+CME ERROR")) {
                /* modem is in weird state, reboot */
                gsm_error(gsm, "error while dialing (%s)\n", buf2);
                gsm->state = GSM_STATE_ERROR;
                gsm->ev.e = GSM_EVENT_ERROR;
                gsm->ev.error.hard = 1;
                gsm->ev.error.cause = 44; /* AST_CAUSE_REQUESTED_CHAN_UNAVAIL */
                strncpy(gsm->ev.error.err, buf2, sizeof(gsm->ev.error.err) - 1);
                return &gsm->ev;
            } else if (gsm_compare(buf2, "NO CARRIER")) { // callee hangup. If a connection cannot be established
                gsm->state = GSM_STATE_READY;
                if (!gsm_compare(gsm->lastcmd, "ATH")) {
                    gsm->ev.e = GSM_EVENT_HANGUP;
                    gsm->ev.hangup.cause = 16;
                    return &gsm->ev;
                }
            } else if (gsm_compare(buf2, "NO DIALTONE")) { // callee hangup. If a connection cannot be established
                gsm->state = GSM_STATE_READY;
                if (!gsm_compare(gsm->lastcmd, "ATH")) {
                    gsm->ev.e = GSM_EVENT_HANGUP;
                    gsm->ev.hangup.cause = 16;
                    return &gsm->ev;
                }
            } else if (gsm_compare(buf2, "NO ANSWER")) {
                gsm->state = GSM_STATE_READY;
                if (!gsm_compare(gsm->lastcmd, "ATH")) {
                    gsm->ev.e = GSM_EVENT_HANGUP;
                    gsm->ev.hangup.cause = 16;
                }
            } else if (gsm_compare(buf2, "MO RING")) {
                gsm->ev.e = GSM_EVENT_ALERTING;
                return &gsm->ev;
            } else if (gsm_compare(buf2, "OK")) {
                if (gsm_compare(gsm->lastcmd, "ATD")) {
                    gsm->lastcmd[0] = '\0';
                } else {
                    gsm->state = GSM_STATE_CALL_ACTIVE;
                    gsm->ev.e = GSM_EVENT_ANSWER;
                    return &gsm->ev;
                }
            } else if (gsm_compare(buf2, "MO CONNECTED")) {
                    gsm->state = GSM_STATE_CALL_ACTIVE;
                    gsm->ev.e = GSM_EVENT_ANSWER;
                    return &gsm->ev;
            } else if (gsm_compare(buf2, "BUSY")) { // if busy
                gsm->state = GSM_STATE_READY;
                gsm->ev.e = GSM_EVENT_HANGUP;
                gsm->ev.hangup.cause = 17;
                return &gsm->ev;
            } else if (gsm_compare(buf2, "ATD")) {
            } else if (gsm_compare(buf2, "+CREG")) {
            } else {
                gsm_error(gsm, DEBUGFMT "!%s!\n", DEBUGARGS, buf2);
            }
            break;
        case GSM_STATE_SM_SEND_REQ:
            if (gsm_compare(buf2, "> ")) {
                m20_transmit(gsm, gsm->msgbuf, strlen(gsm->msgbuf));
                memset(gsm->msgbuf, 0x0, sizeof(gsm->msgbuf));
                snprintf(gsm->msgbuf, sizeof(gsm->msgbuf) - 1, "%c", 0x1A);
                m20_transmit(gsm, gsm->msgbuf, strlen(gsm->msgbuf));
                gsm->state = GSM_STATE_SM_SENDING;
            } else if (gsm_compare(buf2, "+CMS ERROR")) {
                gsm_error(gsm, "error sending SMS on span %d (%s)\n", gsm->span,  buf2);
                gsm->state = GSM_STATE_READY;
            }
            break;
        case GSM_STATE_SM_SENDING:
            if (gsm_compare(buf2, "+CMGS:")) {
                m20_send(gsm, "AT+CMGF=0\r");
                gsm->state = GSM_STATE_READY;
                gsm_message(gsm, "SMS sent successfully on span %d.\n", gsm->span);
            } else if (gsm_compare(buf2, "+CMS ERROR")) {
                gsm_error(gsm, "error sending sms (%s)\n", buf2);
                gsm->state = GSM_STATE_READY;
            }
            break;
        case GSM_STATE_SM_RECEIVING:
            if (!gsm_compare(gsm->lastcmd, "ATE0")) {
                strncpy(gsm->msgbuf, buf2, sizeof(gsm->msgbuf) - 1);
                m20_send(gsm, "ATE0\r");
            }
            gsm->lastcmd[0] = '\0';
            gsm->sanidx = 0;
            memset(gsm->sanbuf, 0, 4096);
            gsm->state = GSM_STATE_READY;
            gsm->ev.e = GSM_EVENT_SM_RECEIVED;
            if (!gsm_pdu2sm_event(gsm, gsm->msgbuf)) {
                return &gsm->ev;
            }
            if (gsm_compare(buf2, "ERROR") || gsm_compare(buf2, "+CMS ERROR")){ // ERROR
                m20_start(gsm);
            }
            break;
        case GSM_STATE_HANGUP_REQ:
            if (gsm_compare(buf2, "OK")){
                gsm->state = GSM_STATE_READY;
                if (gsm->restart_timer) {
                    gsm_schedule_del(gsm, gsm->restart_timer);
                    gsm->restart_timer = 0;
                }
            }
            break;
        case GSM_STATE_STATUS_REQ:
            if (gsm_compare(buf2, "+CSQ:")){
                if (gsm->restart_timer) {
                    gsm_schedule_del(gsm, gsm->restart_timer);
                    gsm->restart_timer = 0;
                }
                gsm->state = GSM_STATE_READY;
                gsm_message(gsm, "signal strength on span %d (%s) [min 0, max 31]\n", gsm->span, buf2);
            }
            break;
    }

    if ((gsm_compare(buf2, "NO CARRIER")) && (gsm->state >  GSM_STATE_READY)){
        gsm->state = GSM_STATE_READY;
        gsm->ev.e = GSM_EVENT_HANGUP;
        gsm->ev.hangup.cause = 16;
        return &gsm->ev;
    }

    if ((gsm_compare(buf2, "NO DIALTONE")) && (gsm->state >  GSM_STATE_READY)){
        gsm->state = GSM_STATE_READY;
        gsm->ev.e = GSM_EVENT_HANGUP;
        gsm->ev.hangup.cause = 16;
        return &gsm->ev;
    }

    if ((gsm_compare(buf2, "BUSY")) && (gsm->state >  GSM_STATE_READY)){
        gsm->state = GSM_STATE_READY;
        gsm->ev.e = GSM_EVENT_HANGUP;
        gsm->ev.hangup.cause = 16;
        return &gsm->ev;
    }

    return NULL;
}
