/*
* opvxg4xx.c - Zaptel driver for the OpenVox GSM PCI cards
*
* Copyright (C) 2005, 2006 OpenVox Communication HK Co Ltd.
*
* Miao Lin <lin.miao@openvox.cn>
*
* This program is free software and may be modified and
* distributed under the terms of the GNU Public License.
*
*
* Rev History:
*	Rev 0.1  2009-July-1 initial version. 
*   Rev 0.2  2010-06-11
*/ 

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/jiffies.h>
#include <zaptel.h>
#include "opvxg4xx.h"

#ifdef LINUX26
#include <linux/moduleparam.h>
#endif

static int debug = 0;
static int g4_dev_count = 0;
static int g4_spans = 0;
static struct pci_dev *multi_gsm = NULL;
static long baudrate    = 115200;   /* default 115200bps. */
static int adcgain      = -3;       /* adc gain of codec, in db, from -42 to 20, mute=21; */
static int dacgain      = -15;      /* dac gain of codec, in db, from -42 to 20, mute=21; */
static int sidetone     = 0;        /* digital side tone, 0=mute, others can be -3 to -21db, step by -3; */
static int inputgain    = 0;        /* input buffer gain, can be 0, 6, 12,24db. */

static int burn = 0;

static char *modem = "default";
static char* G400P_Name = "G400P";

inline unsigned short swaphl(unsigned short i)
{
    return ((i >>8)  & 0x0ff) | (i << 8);
}

inline void __g4_outl__(g4_card* g4, unsigned int base, unsigned int regno, unsigned int value)
{
    writel(value, ((unsigned char *)g4->pci_io) + base + regno * sizeof(int));
}

inline unsigned int __g4_inl__(g4_card* g4, unsigned int base, unsigned int regno)
{
    return readl(((unsigned char *)g4->pci_io) + base + regno * sizeof(int));
}

inline void __g4_pio0_setreg(g4_card* g4, unsigned int regno, unsigned int value)
{
    __g4_outl__(g4, g4->gpio0.base, regno, value);
}

/* set only 1 bit of gpio, value can be 0 or 1 */
inline void __g4_pio0_setbit(g4_card* g4, unsigned int bit, unsigned int value)
{
    if (value) {
        g4->gpio0.io |= (1u << bit);
    }
    else {
        g4->gpio0.io &= ~(1u << bit);
    }
    __g4_outl__(g4, g4->gpio0.base, 0, g4->gpio0.io);
}

/* get only 1 bit of gpio, return 0 or 1 */
inline unsigned int __g4_pio0_getbit(g4_card* g4, unsigned int bit)
{
    unsigned int i = __g4_inl__(g4, g4->gpio0.base, 0);
    return  (i >> bit) & 1u;
}

/* set more 1 bit of gpio, only masked bit can be set */
inline void __g4_pio0_setbits(g4_card* g4, unsigned int mask, unsigned int value)
{
    g4->gpio0.io &= ~mask;
    g4->gpio0.io |= mask & value;
    __g4_outl__(g4, g4->gpio0.base, 0, g4->gpio0.io);
}

/* get more 1 bit of gpio, only masked bit can be get */
inline unsigned int __g4_pio0_getbits(g4_card* g4, unsigned int mask)
{
    return __g4_inl__(g4, g4->gpio0.base, 0) & mask;
}

/* !!!!!! this function can not be used when irq closed, for example ....irq_save() !*/
void _wait(int msec)
{
    unsigned long origjiffies = jiffies;

    while(1) {
        if (jiffies_to_msecs(jiffies - origjiffies) >= msec) {
            break;
        }
    }
}

void uart_init(g4_card* g4)
{
    unsigned int magic;
    
    switch (baudrate) {
        case 9600:          magic = 6875;           break;
        case 19200:         magic = 3438;           break;
        case 38400:         magic = 1719;           break;
        case 57600:         magic = 1146;           break;
        case 115200:        magic = 573;            break;
        default:            magic = 573;            break;  /* ml we default use 115200 */
    }
    
    __g4_outl__(g4, UART0_BASE, 5, magic);
    __g4_outl__(g4, UART1_BASE, 5, magic);
    __g4_outl__(g4, UART2_BASE, 5, magic);
    __g4_outl__(g4, UART3_BASE, 5, magic);

    /* set work mode : no irq, RTS on ( active low) */
    __g4_outl__(g4, UART0_BASE, 3, 1 << 11);
    __g4_outl__(g4, UART1_BASE, 3, 1 << 11);
    __g4_outl__(g4, UART2_BASE, 3, 1 << 11);
    __g4_outl__(g4, UART3_BASE, 3, 1 << 11);
}

/* G400 codec specific functions. */
inline unsigned char format_adcgain(int gain)
{
    unsigned char ret = 0x2a;   /* default adcgain=0db */
    
    if((gain > 21) || (gain < -42)) {
        printk(KERN_WARNING "opvxg4xx: Invalid adcgain %d\n", gain);
    }
    else {
        int i = 0x2a;
        if(gain == 21) {
            ret = 0x3f; /* Mute. */
        }
        else {
                ret = (unsigned char)(i + gain);    /* change to correct format; */
        }
    }

    if(debug & DEBUG_G4_CODEC) {
        printk(KERN_DEBUG "opvxg4xx: adcgain is 0x%x(%ddb)\n", ret & 0xff, gain);
    }
    
    return ret;
}

inline unsigned char format_dacgain(int gain)
{
    unsigned char ret = 0x56;   /* default dacgain=-20db */
    
    if(gain > 21 || gain < -42) {
        printk(KERN_WARNING "opvxg4xx: Invalid dacgain %d\n", gain);
    }
    else {
        int i = 0x6a;
        if(gain == 21) {
            ret = 0x7f; /* Mute. */
        }
        else {
            ret = (unsigned char)(i + gain);    /* change to correct format; */
        }
    }

    if(debug & DEBUG_G4_CODEC) {
        printk(KERN_DEBUG "opvxg4xx: dacgain is 0x%x(%ddb)\n", ret & 0xff, gain);
    }
    
    return ret;
}

inline unsigned char format_sidetone(int sidetone)
{
    unsigned char ret = 0x7;    /* default mute. */
    int i;
    int valid_values[] = {0, -3, -6, -9, -12, -15, -18, -21};
    int num_values = sizeof(valid_values) / sizeof(valid_values[0]);
    
    for (i = 0; i < num_values; i ++) {
        if(sidetone == valid_values[i]) {
            break;
        }
    }
        
    if (i < num_values) {
        if(i == 0){ /* set to mute; */
            if (debug & DEBUG_G4_CODEC) {
                printk(KERN_DEBUG "opvxg4xx: sidetone set to mute\n");
            }
        }
        else {
            ret = i-1;
            if (debug & DEBUG_G4_CODEC) {
                printk(KERN_DEBUG "opvxg4xx: sidetone set to %ddb\n", valid_values[i]);
            }
        }
    }
    else {
        printk(KERN_WARNING "opvxg4xx: sidetone value %d not correct, set to default 7(MUTE)\n", sidetone);
    }
    
    return ret;
}

inline unsigned char format_inputgain(int gain)
{
    unsigned char ret = 0;  /* default 0db. */
    int i;
    int valid_values[] = {0, 6, 12, 24 };
    int num_values = sizeof(valid_values) / sizeof(valid_values[0]);

    for(i = 0; i < num_values; i ++) {
        if(gain == valid_values[i]) {
            break;
        }
    }

    if(i < num_values) {
        ret = i;
        if(debug & DEBUG_G4_CODEC) {
            printk(KERN_DEBUG "opvxg4xx: input gain set to %ddb\n", valid_values[i]);
        }
    }
    else {
        printk(KERN_WARNING "opvxg4xx: input gain value %d not correct, set to default 0db\n", gain);
    }
    
    return ret;
}

void pio0_init(g4_card* g4)
{
    /* all uart signals are low level enable, and power is also reversed */
    g4->gpio0.base  = PIO0_BASE;
    g4->gpio0.dir   = 0x0fff000f;   /* set io direction */
    g4->gpio0.io    = 0x000f0000;   /* set value: dtr 0, power 0, rst 1, portsel 0 */

    __g4_outl__(g4, g4->gpio0.base, 1, g4->gpio0.dir);
    __g4_outl__(g4, g4->gpio0.base, 2, 0);         /* set interrupt mask to 0 */
    __g4_outl__(g4, g4->gpio0.base, 3, 0);         /* no edge capture */
    __g4_outl__(g4, g4->gpio0.base, 0, g4->gpio0.io);   
}

/* G400 i2c specific functions. */
inline void i2c_Init(unsigned int *_base)
{
    volatile unsigned int *base = _base;

    /* prescale = busfreq/(5*I2cFreq), now we use 100k I2c Freq; */
    unsigned prescale = BUS_FREQ/5/100000;

    base[I2C_FREQ_LOW] = prescale&0xff;
    base[I2C_FREQ_HIGH] = (prescale>>8)&0xff;
    base[I2C_CTRL] = I2C_CTRL_ENABLE;   /* we do not use irq. */
}

int i2c_SetRead(unsigned int *_base, unsigned char addr)
{
    volatile unsigned int *base = _base;
    unsigned long origjiffies;

    base[I2C_TX] = 0x80; /*I2C address is 0x100, SamrtPCM address is 0, I2C WR; */
    base[I2C_CMD] = I2C_CMD_STA | I2C_CMD_WR;
    /* while(base[I2C_STATUS] & I2C_STATUS_TIP){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_TIP)) {
            break;
        }
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -1;
        }
    }

    /* while(base[I2C_STATUS] & I2C_STATUS_RXACK){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_RXACK)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -2;
        }
    }

    _wait(10);
    base[I2C_TX] = addr;
    base[I2C_CMD] = I2C_CMD_STO | I2C_CMD_WR;
    /* while(base[I2C_STATUS] & I2C_STATUS_TIP){}; */
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_TIP)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -3;
        }
    }

    /* while(base[I2C_STATUS] & I2C_STATUS_RXACK){}; */
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_RXACK)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -4;
        }
    }

    return 0;
}

int i2c_Read(unsigned int *_base, unsigned char* pVal)
{
    volatile unsigned int *base = _base;
    unsigned long origjiffies;

    base[I2C_TX] = 0x81; /* I2C address is 0x100, SamrtPCM address is 0, I2c Read; */
    base[I2C_CMD] = I2C_CMD_STA | I2C_CMD_WR;

    /* while(base[I2C_STATUS] & I2C_STATUS_TIP){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_TIP)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -1;
        }
    }

    base[I2C_CMD] = I2C_CMD_RD | I2C_CMD_ACK | I2C_CMD_STO;
    /* while(base[I2C_STATUS] & I2C_STATUS_TIP){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_TIP)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -2;
        }
    }
    *pVal = base[I2C_RX];

    return 0;
}

int i2c_Write(unsigned int *_base, unsigned char reg, unsigned char value)
{
    unsigned long origjiffies;
    volatile unsigned int *base = _base;

    base[I2C_TX] = 0x80;    /*I2C address is 0x100, SamrtPCM address is 0, I2C WR; */
    base[I2C_CMD] = I2C_CMD_STA | I2C_CMD_WR;
    /* while(base[I2C_STATUS] & I2C_STATUS_TIP){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_TIP)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -1;
        }
    }

    /* while(base[I2C_STATUS] & I2C_STATUS_RXACK){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_RXACK)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -2;
        }
    }

    base[I2C_TX] = reg; /* I2C address is 0x100, SamrtPCM address is 0, I2C WR; */
    base[I2C_CMD] = I2C_CMD_WR;
    /* while(base[I2C_STATUS] & I2C_STATUS_TIP){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_TIP)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -3;
        }
    }

    /* while(base[I2C_STATUS] & I2C_STATUS_RXACK){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_RXACK)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -4;
        }
    }

    base[I2C_TX] = value;   /* I2C address is 0x100, SamrtPCM address is 0, I2C WR; */
    base[I2C_CMD] = I2C_CMD_STO | I2C_CMD_WR;
    /* while(base[I2C_STATUS] & I2C_STATUS_TIP){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_TIP)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -5;
        }
    }

    /* while(base[I2C_STATUS] & I2C_STATUS_RXACK){}; */
    origjiffies = jiffies;
    while(1) {
        if(!(base[I2C_STATUS] & I2C_STATUS_RXACK)) {
            break;
        }
        
        if (jiffies_to_msecs(jiffies - origjiffies) >= I2C_TIMEOUT) {
            return -6;
        }
    }

    return 0;
}

void init_320aic14k(unsigned int *base)
{
    unsigned char c;
    int i;

    i = i2c_SetRead(base, 2);
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }
    i = i2c_Read(base, &c);
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }
    i = i2c_Write(base, 2, c|0x80); /* set turbo mode */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }
    i = i2c_Write(base, 1, 0x41);   /* cx=1, dac16=1, continuous data transfer, 16bit DAC input */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }

    i = i2c_Write(base, 3, 0x01);   /* FS/Fs=1, OSR for DAC channels is 128 */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }

    /* set m=1, n=2, p=8(default); */
    i = i2c_Write(base, 4, 0x88);   /* set m=8, ( FSDIV=1, D6-0=8) */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }
    i = i2c_Write(base, 4, 0x01);   /* set n=16, p=1, ( FSDIV=0); */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }
    /*  A/D converter gain */
    c = format_adcgain(adcgain);
    i = i2c_Write(base, 5, c);       /* setup adc gain, reg5a; */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }

    c = format_dacgain(dacgain);
    i = i2c_Write(base, 5, c);    /* setup dac gain, reg5b; */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }

    c = 0x80;
    c |= (format_sidetone(sidetone) << 3) | format_inputgain(inputgain);
    i = i2c_Write(base, 5, c);       /* setup sidetone gain and input buffer gain, reg 5c; */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }

    i = i2c_Write(base, 5, 0xff);   /* useless, reg5d. */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }

    i = i2c_Write(base, 6, 0xb0);   /* disable out2/3 */
    if(debug & DEBUG_G4_I2C) {
        printk(KERN_DEBUG "initaic14k return %x\n", i);
    }
}

inline void g4_power_on(struct g4_card *g4, int delay)
{
    unsigned long flags;
	
    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((3000 * HZ) / 1000);
	
    spin_lock_irqsave(&(g4->lock),flags);
    __g4_pio0_setbits(g4, PIO0_PWR_MASK, PIO0_PWR_MASK);  /* set PWR_N to low for 2500 ms */
    spin_unlock_irqrestore(&(g4->lock),flags);

    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((delay * HZ) / 1000);

    spin_lock_irqsave(&(g4->lock),flags);
    __g4_pio0_setbits(g4, PIO0_DTR_MASK, 0x0);  /* set DTR to low, Enable DTR pin to wake up SIM300DZ. If DTR Pin is pull down to a low level.*/
    __g4_pio0_setbits(g4, PIO0_PWR_MASK, 0x0);  /* set PWR_N to high */
    __g4_pio0_setbits(g4, PIO0_RST_MASK, 0x0);  /* set RST to low for 100 ms */
    spin_unlock_irqrestore(&(g4->lock),flags);

    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((100 * HZ) / 1000);
    
    spin_lock_irqsave(&(g4->lock),flags);
    __g4_pio0_setbits(g4, PIO0_RST_MASK, PIO0_RST_MASK);  /* set RST to HIGH*/
    g4->power[0] = 1;
    g4->power[1] = 1;
    g4->power[2] = 1;
    g4->power[3] = 1;
    spin_unlock_irqrestore(&(g4->lock),flags);	

    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((3000 * HZ) / 1000);
	
}

inline void g4_power_off(struct g4_card *g4, int delay)
{
    unsigned long flags;
    	
    spin_lock_irqsave(&(g4->lock),flags);
    __g4_pio0_setbits(g4, PIO0_PWR_MASK, PIO0_PWR_MASK);  /* set PWR_N to low */
    spin_unlock_irqrestore(&(g4->lock),flags);

    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((delay * HZ) / 1000);
    
    spin_lock_irqsave(&(g4->lock),flags);
    __g4_pio0_setbits(g4, PIO0_PWR_MASK, 0x0);  /* set PWR_N to high */
    g4->power[0] = 0;
    g4->power[1] = 0;
    g4->power[2] = 0;
    g4->power[3] = 0;
    spin_unlock_irqrestore(&(g4->lock),flags);
    
    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((3000 * HZ) / 1000);

    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout((3000 * HZ) / 1000);    
}

void g4_switch_on_all(struct g4_card *g4, int verbose)
{
    unsigned long flags;
    
    if (verbose) {
        printk(KERN_INFO "opvxg4xx: Powering up all spans...\n");
    }

    if (g4->rst_flag || 1 == burn) {
        spin_lock_irqsave(&(g4->lock),flags);
        __g4_pio0_setbits(g4, PIO0_DTR_MASK, 0x0);              /* set DTR to low, Enable DTR pin to wake up SIM300DZ. If DTR Pin is pull down to a low level.*/
        __g4_pio0_setbits(g4, PIO0_RST_MASK, 0x0);  /* set RST to low for 100 ms */
        spin_unlock_irqrestore(&(g4->lock),flags);

        set_current_state(TASK_UNINTERRUPTIBLE);
        schedule_timeout((100 * HZ) / 1000);
        
        spin_lock_irqsave(&(g4->lock),flags);
        __g4_pio0_setbits(g4, PIO0_RST_MASK, PIO0_RST_MASK);    /* set RST to HIGH*/        
        g4->power[0] = 1;
        g4->power[1] = 1;
        g4->power[2] = 1;
        g4->power[3] = 1;
        spin_unlock_irqrestore(&(g4->lock),flags);
        return;
    } else {
        spin_lock_irqsave(&(g4->lock),flags);
        __g4_outl__(g4, SWAP_MEM_BASE, 3, 1);
        g4->rst_flag = 1;
        spin_unlock_irqrestore(&(g4->lock),flags);
        if (!strcmp(modem, "default")) {
            g4_power_on(g4, 2500);
        } else if (!strcmp(modem, "sim340dz")) {
            g4_power_on(g4, 2500);
        } else if (!strcmp(modem, "em200")) {
            g4_power_on(g4, 100);
        } else if (!strcmp(modem, "m20")) {
            g4_power_on(g4, 2000);
        } else {
            printk(KERN_ERR "opvxg4xx: modem parameter error!");
        }
    }
}

void g4_switch_off_all(struct g4_card *g4, int verbose)
{
    unsigned long flags;
    
    spin_lock_irqsave(&(g4->lock),flags);
    __g4_outl__(g4, SWAP_MEM_BASE, 3, 0);
    g4->rst_flag = 0;
    spin_unlock_irqrestore(&(g4->lock),flags);

    if (g4->power[0] || g4->power[1] || g4->power[2] || g4->power[3]) {
        if (verbose) {
            printk(KERN_INFO "opvxg4xx: Powering down all spans...\n");
        }

        if (1 == burn) {
            return;
        }
        
        if (!strcmp(modem, "default")) {
            g4_power_off(g4, 2500);
        } else if (!strcmp(modem, "sim340dz")) {
            g4_power_off(g4, 1500);
        } else if (!strcmp(modem, "em200")) {
            g4_power_off(g4, 2500);
        } else if (!strcmp(modem, "m20")) {
            g4_power_off(g4, 800);
        } else {
            printk(KERN_ERR "opvxg4xx: modem parameter error!");
        }

        if (verbose) {
            printk(KERN_INFO " done.\n");
        }
        
    }
}

void g4_hardware_init(struct g4_card *g4)
{
    unsigned long flags;
    spin_lock_irqsave(&(g4->lock),flags); /*change to spin_lock, due to jiffies can not work when disable irq;*/

    uart_init(g4);

    pio0_init(g4);


    spin_unlock_irqrestore(&(g4->lock),flags);
}

void g4_startCard(struct g4_card *g4)
{
    unsigned long flags;
    unsigned int err;

    spin_lock_irqsave(&(g4->lock), flags);
    __g4_outl__(g4, SWAP_MEM_BASE, 0, 1);                        /* start interrupts; */
    err = __g4_inl__(g4, SWAP_MEM_BASE, 1);
    if (debug & DEBUG_G4_SYS) {
        printk(KERN_ERR "err = %d\n", err);
    }

    __g4_outl__(g4, IRQ_BASE, 0x0040/sizeof(int), 0xffffffff);   /* clear all pending interrupts; */
    __g4_outl__(g4, IRQ_BASE, 0x0050/sizeof(int), 1<<16);        /* enable interrupt enable reg; */
    spin_unlock_irqrestore(&(g4->lock),flags);
    return;
}

static int g4_ser_rx(struct g4_card *g4, int span)
{
    unsigned int uart_base = 0;
    int rx_count = 0;
    int i = 0;
    unsigned char data = 0;
    unsigned int count_reg = 0;

    switch (span) {
        case 0: uart_base = UART0_BASE; break;
        case 1: uart_base = UART1_BASE; break;
        case 2: uart_base = UART2_BASE; break;
        case 3: uart_base = UART3_BASE; break;
    }

    count_reg = __g4_inl__(g4, uart_base, 6);

    rx_count = count_reg;
    if (rx_count) {
        if ((g4->ser_rx_idx[span] + rx_count) <= G4_SER_BUF_SIZE) {
            if (debug & DEBUG_G4_SER) {
                printk(KERN_INFO "opvxg4xx: span %d SER_RX [ ", span);
            }

            for (i = 0; i < rx_count; i ++) {
                data = __g4_inl__(g4, uart_base, 0);
                if (debug & DEBUG_G4_SER) {
                    if (data == 0xd) {
                        printk("\\r");
                    } else if (data == 0xa) {
                        printk("\\n");
                    } else {
                        printk("%c", data);
                    }
                }
                if (g4->spans[span].span.flags & ZT_FLAG_RUNNING) {
                    g4->ser_rx_buf[span][g4->ser_rx_idx[span]++] = data;
                }
            }

            if (debug & DEBUG_G4_SER) {
                printk(" ]\n");
            }
        } else {
                printk(KERN_WARNING "opvxg4xx: RX buffer overflow on span %d, idx %d, cnt %d, size %d\n", 
                            span, g4->ser_rx_idx[span], rx_count, G4_SER_BUF_SIZE);
        }
    }

    return rx_count;
}

static int g4_ser_tx(struct g4_card *g4, int span)
{
    unsigned int uart_base = 0;
    int left = 0;
    int i = 0;
    int count = 0;
    unsigned int count_reg = 0;
    int rd_ptr = 0;
    int wr_ptr = 0;
    struct g4_span *myspan = NULL;

    switch (span) {
        case 0: uart_base = UART0_BASE; break;
        case 1: uart_base = UART1_BASE; break;
        case 2: uart_base = UART2_BASE; break;
        case 3: uart_base = UART3_BASE; break;
    }

    myspan = &g4->spans[span];
    if (myspan) {
        count_reg = __g4_inl__(g4, uart_base, 7);
        left = G4_FIFO_SIZE - (count_reg & 0x1F); /*  this code should be modified future, since we have 1024bytes fifo. ml */
        if (left >= 1 ) {
            if (debug > DEBUG_G4_SER) {
                printk(KERN_CRIT "opvxg4xx: SER_TX_COUNT_%d [wr_ptr %d rd_ptr %d free %d]\n", span, wr_ptr, rd_ptr, left);
            }
            if (g4->ser_tx_idx[span] < left) {
                count = g4->ser_tx_idx[span];
            } else {
                count = left;
            }
            
            if (debug & DEBUG_G4_SER) {
                printk(KERN_INFO "opvxg4xx: span %d SER_TX [ ", span);
            }

            for (i = 0; i < count; i++) {
                if ((debug & DEBUG_G4_SER) && (g4->ser_tx_buf[span][i] > 0x10))
                    printk("%c", g4->ser_tx_buf[span][i]);
                __g4_outl__(g4, uart_base, 1, g4->ser_tx_buf[span][i]);
            }
            
            if (debug & DEBUG_G4_SER) {
                printk(" ]\n");
            }
            
            g4->ser_tx_idx[span] -= count;

            if (g4->ser_tx_idx[span] > 0) {
                memmove(&g4->ser_tx_buf[span][0], &g4->ser_tx_buf[span][i], g4->ser_tx_idx[span]);
            }
        } else {
            if (debug & DEBUG_G4_SER) {
                printk(KERN_DEBUG "opvxg4xx: span %d wanted to TX %d bytes but only %d bytes free fifo space\n", span, g4->ser_tx_idx[span], left);
            }
        }
    }
    
    return i;
}

static int g4_zap_rx(struct g4_card *g4, int span)
{
    int i = 0;
    struct g4_span *myspan = NULL;

    myspan = &g4->spans[span];
    if (!myspan)
        return -1;

    if ((g4->ser_rx_idx[span]) && (!g4->d_rx_recv[span])) {
        memcpy(g4->d_rx_buf[span], g4->ser_rx_buf[span], g4->ser_rx_idx[span]);
        if (debug & DEBUG_G4_ZAP) { 
            printk(KERN_CRIT "opvxg4xx: span %d RX [ ", span);
            for (i = 0; i < g4->ser_rx_idx[span]; i++) {
                if (g4->ser_rx_buf[span][i] == 0xd) {
                    printk("\\r");
                } else if (g4->ser_rx_buf[span][i] == 0xa) {
                    printk("\\n");
                } else if (g4->ser_rx_buf[span][i] == 0x0) {
                    printk("NULL");
                } else {
                    printk("%c", g4->ser_rx_buf[span][i]);
                }
            }
            printk(" ] %d \n", g4->ser_rx_idx[span]);
        }

        g4->spans[span].chans[1].eofrx = 1;
        g4->spans[span].chans[1].bytes2receive = g4->ser_rx_idx[span];
        g4->ser_rx_idx[span] = 0;
    }

    return 0;
}

static int g4_zap_tx(struct g4_card *g4, int span)
{
    struct g4_span *myspan = NULL;
    int i = 0;
    
    myspan = &g4->spans[span];

    if (!myspan)
        return -1;

    if (myspan->chans[1].bytes2transmit) {
        if (debug & DEBUG_G4_ZAP) {
            printk("opvxg4xx: span %d TX [ ", span);
            for (i = 0; i < myspan->chans[1].bytes2transmit; i++) {
                if (g4->d_tx_buf[span][i] > 0x10)
                    printk("%c", g4->d_tx_buf[span][i]);
            }
            printk(" ]\n");
        }
        
        if (g4->power[span]) {
            if (g4->ser_tx_idx[span] + myspan->chans[1].bytes2transmit < G4_SER_BUF_SIZE) {
                memcpy(&g4->ser_tx_buf[span][g4->ser_tx_idx[span]], g4->d_tx_buf[span], myspan->chans[1].bytes2transmit);
                g4->ser_tx_idx[span] += myspan->chans[1].bytes2transmit;
                g4_ser_tx(g4, span);
            } else {
                printk(KERN_WARNING "opvxg4xx: TX buffer overflow on span %d (TX_IDX %d BTT %d)\n", span, g4->ser_tx_idx[span] , myspan->chans[1].bytes2transmit);
            }
        }
    }

    myspan->chans[1].bytes2receive = 0;
    myspan->chans[1].bytes2transmit = 0;
    myspan->chans[1].eofrx = 0;
    myspan->chans[1].eoftx = 0;

    return 0;
}

static int g4_bchan_rx(struct g4_card *g4, int span)
{
    struct g4_span *myspan = NULL;

    myspan = &g4->spans[span];
    if (myspan) {
        volatile unsigned short* p;
        unsigned int i, j;
        
        j = __g4_inl__(g4, TDM0_BASE, 8);
        if(j & 0x8) /* tdm controller is using high helf, pcm to mem is located at 1st half. */
            p = (unsigned short*) ((unsigned char*)(g4->pci_io) + DP_MEM_BASE);
        else
            p = (unsigned short*) ((unsigned char*)(g4->pci_io) + DP_MEM_BASE + 8*128);
        
        for(i = 0; i < ZT_CHUNKSIZE; i++) {
            unsigned short line = p[i*64+span];
            g4->b_rx_buf[span][i] = ZT_LIN2X(swaphl(line), myspan->chans);
        }
    }

    return 0;
}

static int g4_bchan_tx(struct g4_card *g4, int span) 
{
    struct g4_span *myspan = NULL;

    myspan = &g4->spans[span];
    if (myspan) {
        volatile unsigned short* p;
        unsigned int i, j;
        
        j = __g4_inl__(g4, TDM0_BASE, 8);
        if(j & 0x8) /* tdm controller is using high helf, mem to pcm is located at 2nd half. */
            p = (unsigned short*) ((unsigned char*)(g4->pci_io) + DP_MEM_BASE + 2048);
        else
            p = (unsigned short*) ((unsigned char*)(g4->pci_io) + DP_MEM_BASE + 2048 + 8*128);
        
        for(i = 0; i < ZT_CHUNKSIZE; i++) {
            unsigned short line = ZT_XLAW(g4->b_tx_buf[span][i], myspan->chans);
            p[i*64+span] = swaphl(line);
        }
    }
    return 0;
}

static inline void g4_run(struct g4_card *g4)
{
    int s=0;
    struct g4_span *myspan = NULL;

    for (s=0;s<g4->g4spans;s++) {
        myspan = &g4->spans[s];
        if (myspan) {
            if (myspan->span.flags & ZT_FLAG_RUNNING) {
                zt_transmit(&myspan->span);
                /*
                if (debug && (gsmspan->chans[1].bytes2transmit > 0)) {
                    printk(KERN_CRIT "opvxg4xx: span %d bytes2transmit = %d\n", s, g4->chans[1].bytes2transmit);
                } */
                g4_bchan_tx(g4, s);
                g4_zap_tx(g4, s);
            }
            
            if (myspan->span.flags & ZT_FLAG_RUNNING) {
                g4_zap_rx(g4, s);
                g4_bchan_rx(g4, s);
                zt_receive(&myspan->span);
            }
        } 
    }
}



ZAP_IRQ_HANDLER(g4_interrupt)
{
    struct g4_card *g4 = dev_id;
    int s = 0;
    unsigned int i;

    if (!g4 || g4->dead) {
#ifdef LINUX26
        return IRQ_NONE;
#else
        return;
#endif      
    }

    if (!g4->pci_io ) {
        printk(KERN_CRIT "opvxg4xx: no pci mem/io\n");
#ifdef LINUX26
        return IRQ_NONE;
#else
        return;
#endif      
    }

    /* simply clear the irq and return, for test. ml */
    i = __g4_inl__(g4, IRQ_BASE, 0x0040/sizeof(int));
    if(i & ( 1<<16)) {
        __g4_outl__(g4, IRQ_BASE, 0x0040/sizeof(int), i); /* simply clear it; */
    }
    else {
#ifdef LINUX26
        return IRQ_NONE;
#else
    return;
#endif      
    }

    g4_run(g4); /* realwork in this function;     */

    for (s = 0; s < g4->g4spans; s++) {
        g4->d_rx_recv[s] = g4_ser_rx(g4, s);
        if (g4->ser_tx_idx[s]) {
            g4_ser_tx(g4, s); /* sumfin left to transmit */
        }
    }

#ifdef LINUX26
    return IRQ_RETVAL(1);
#endif
}


static int g4_open(struct zt_chan *chan)
{
#ifndef LINUX26
    MOD_INC_USE_COUNT;
#else
    try_module_get(THIS_MODULE);
#endif
    return 0;
}

static int g4_close(struct zt_chan *chan)
{
#ifndef LINUX26
    MOD_DEC_USE_COUNT;
#else
    module_put(THIS_MODULE);
#endif
    return 0;
}

static int g4_chanconfig(struct zt_chan *chan,int sigtype)
{
    return 0;
}

static int g4_spanconfig(struct zt_span *span,struct zt_lineconfig *lc)
{
    return 0;
}

static int g4_startup(struct zt_span *span)
{
    struct g4_card *g4 = span->pvt;
    int running;

    if (g4 == NULL) {
        printk(KERN_INFO "opvxg4xx: no card for span at startup!\n");
    }

    running = span->flags & ZT_FLAG_RUNNING;

    if (!running) {
        span->chans[1].flags &= ~ZT_FLAG_HDLC;
        span->chans[1].flags |= ZT_FLAG_BRIDCHAN;
        if (!g4->power[span->offset]) 
        {
            printk(KERN_WARNING "g4_switch_on must be called, but now it still not ready!\n");
//          g4_switch_on(g4, span->offset);
        }
        span->flags |= ZT_FLAG_RUNNING;
    } else {
        printk(KERN_INFO "opvxg4xx: already running\n");
        return 0;
    }
    
    return 0;
}

static int g4_shutdown(struct zt_span *span)
{
    int running;
    struct g4_card *g4 = span->pvt;

    if (g4 == NULL)
        printk(KERN_INFO "opvxg4xx: no card for span at shutdown!\n");

    running = span->flags & ZT_FLAG_RUNNING;
    
    if (running) {
        span->flags &= ~ZT_FLAG_RUNNING;
        if (g4->power[span->offset]) {
//          g4_switch_off(g4, span->offset);
        }
    }

    return 0;
}

static int g4_ioctl(struct zt_chan *chan, unsigned int cmd, unsigned long data)
{
    switch(cmd) {
    default:
        return -ENOTTY;
    }
    return 0;
}


static int g4_init_span(struct g4_span *myspan, struct g4_card *g4, int offset)
{
    memset(&myspan->span,0,sizeof(struct zt_span));
    sprintf(myspan->span.name,"opvxg4xx/%d/%d",g4->cardno, offset + 1);
    switch (g4->type) {
    case 0x0104:
        sprintf(myspan->span.desc,"OpenVox G400P GSM/CDMA PCI Card %d", g4->cardno);    /* we always be 4 ports card */
        break;
    }
        
    myspan->span.spanconfig = g4_spanconfig;
    myspan->span.chanconfig = g4_chanconfig;
    myspan->span.startup = g4_startup;
    myspan->span.shutdown = g4_shutdown;
    myspan->span.maint = NULL;
    myspan->span.rbsbits = NULL;
    myspan->span.irq = g4->pcidev->irq; 
    myspan->span.open = g4_open;
    myspan->span.close = g4_close;
    myspan->span.ioctl = g4_ioctl;
    myspan->span.flags = 0;
    myspan->span.chans = myspan->chans;
    myspan->span.channels = 2;
    myspan->span.deflaw = ZT_LAW_ALAW;
    myspan->span.linecompat = ZT_CONFIG_CCS | ZT_CONFIG_AMI; 
    init_waitqueue_head(&myspan->span.maintq);
    myspan->span.pvt = g4;
    myspan->span.offset = offset;

    memset(&(myspan->chans[0]),0x0,sizeof(struct zt_chan));
    sprintf(myspan->chans[0].name,"opvxg4xx/%d/%d/%d", g4->cardno, offset + 1, 0);
    myspan->chans[0].pvt = myspan;
    myspan->chans[0].sigcap =  ZT_SIG_CLEAR;
    myspan->chans[0].chanpos = 1; 

    memset(&(myspan->chans[1]),0x0,sizeof(struct zt_chan));
    sprintf(myspan->chans[1].name,"opvxg4xx/%d/%d/%d", g4->cardno, offset + 1, 1);
    myspan->chans[1].pvt = myspan;
    myspan->chans[1].sigcap =  ZT_SIG_CLEAR;
    myspan->chans[1].chanpos = 2;


    if (zt_register(&myspan->span,0)) {
        printk(KERN_INFO "opvxg4xx: unable to register dahdi span!\n");
        return -1;
    }

    /* setup B channel buffers (8 bytes each) */
    memset(g4->b_rx_buf[offset],0x0,sizeof(g4->b_rx_buf[offset]));
    myspan->span.chans[0].readchunk = g4->b_rx_buf[offset];          /* readchunk */
    memset(g4->b_tx_buf[offset],0x0,sizeof(g4->b_tx_buf[offset]));
    myspan->span.chans[0].writechunk = g4->b_tx_buf[offset];         /* writechunk */

    /* setup D channel buffer */
    memset(g4->d_rx_buf[offset],0x0,sizeof(g4->d_rx_buf[offset]));
    myspan->span.chans[1].readchunk = g4->d_rx_buf[offset];          /* readchunk */
    memset(g4->d_tx_buf[offset],0x0,sizeof(g4->d_tx_buf[offset]));
    myspan->span.chans[1].writechunk = g4->d_tx_buf[offset];         /* writechunk */
    myspan->span.chans[1].maxbytes2transmit = sizeof(g4->d_tx_buf[offset]); /* maxbytes2transmit */


    return 0;
}

static int __devinit g4_probe(struct pci_dev *tmp, const struct pci_device_id *ent)
    {
    struct g4_card *g4 = NULL;
    struct g4_span *myspan = NULL;
    int i = 0;
    unsigned long status;

    multi_gsm = tmp;
    
    if (pci_enable_device(tmp)) {
        return -1;
    }
        
    g4 = kmalloc(sizeof(struct g4_card),GFP_KERNEL);
    if (!g4) {
        printk(KERN_WARNING "opvxg4xx: unable to kmalloc!\n");
        pci_disable_device(tmp);
        return -ENOMEM;
    }
    memset(g4, 0x0, sizeof(struct g4_card));

    spin_lock_init(&g4->lock);
    g4->pcidev = tmp;

    if (!tmp->irq) {
        printk(KERN_WARNING "opvxg4xx: no irq!\n");
    } else {
        g4->irq = tmp->irq;
    }

    if (debug & DEBUG_G4_SYS) {   
        printk(KERN_DEBUG "bus number = %d devfn = %d irq = %d\n", 
            tmp->bus->number, tmp->devfn, g4->irq);
    }

    /* We only have 1 64KB memory region. */
    g4->pci_io_phys = tmp->resource[0].start;
    if (debug & DEBUG_G4_SYS) {   
        printk(KERN_DEBUG "pci-io-phys = %lx\n", g4->pci_io_phys);
    }
    if (!g4->pci_io_phys) {
        printk(KERN_WARNING "opvxg4xx: no iomem!\n");
        pci_disable_device(tmp);
        return -EIO;
    }

    g4->iomem_size = (tmp->resource[0].end - tmp->resource[0].start + 1);   /* our region is at bar0 */
    if (debug & DEBUG_G4_SYS) {
        printk(KERN_DEBUG "opvxg4xx: iomem at %lx size %ld\n", g4->pci_io_phys, g4->iomem_size);
    }

    if (check_mem_region(g4->pci_io_phys, g4->iomem_size)) {
        printk(KERN_WARNING "opvxg4xx: iomem already in use!\n");;
        pci_disable_device(tmp);
        return -EBUSY;
    }

    request_mem_region(g4->pci_io_phys, g4->iomem_size, "opvxg4xx");
    g4->pci_io = ioremap(g4->pci_io_phys, g4->iomem_size); /* 64kb */
    if (debug & DEBUG_G4_SYS) {
        printk(KERN_DEBUG "opvxg4xx: iomem remote to %lx \n", (unsigned long)g4->pci_io);
    }

    if (request_irq(g4->irq, g4_interrupt, ZAP_IRQ_SHARED, "opvxg4xx", g4)) {
        printk(KERN_WARNING "opvxg4xx: unable to register irq\n");
        release_mem_region(g4->pci_io_phys, g4->iomem_size);
        kfree(g4);
        pci_disable_device(tmp);
        return -EIO;
    }

    
    g4->fwversion = __g4_inl__(g4, SWAP_MEM_BASE, 2);
    g4->card_name = G400P_Name;
    g4->rst_flag = __g4_inl__(g4, SWAP_MEM_BASE, 3);
    
    g4->type = tmp->subsystem_device;
    switch (g4->type) {
        case PCI_SUBSYSTEM_ID_OPENVOX:
            printk(KERN_NOTICE
                "Found an OpenVox %s: Version %x.%x\n", g4->card_name, g4->fwversion>>16, g4->fwversion&0xffff);
            break;
    }
    pci_set_drvdata(tmp, g4);

    g4_hardware_init(g4);

    status = __g4_pio0_getbits(g4, PIO0_STATUS_MASK | PIO0_SLOTUSE_MASK);
    for (i = 0; i < 4; i ++) {
        printk(KERN_NOTICE "opvxg4xx: slot %d is %s\n", i, status&(1<<(PIO0_SLOTUSE+i)) ? "Empty" : "Installed" );
    }
    
    g4->g4spans = 4;    /* we always have 4 spans */
    g4_spans += g4->g4spans;

    g4->cardno = g4_dev_count ++;
    for (i = 0; i < g4->g4spans; i++) {
        myspan = &g4->spans[i];
        g4_init_span(myspan, g4, i);
    }

    g4_startCard(g4);

    g4_switch_on_all(g4, 1);
    
    if (debug & DEBUG_G4_SYS) {
        printk(KERN_DEBUG "pci_io addr is 0x%lx \n", (unsigned long)g4->pci_io);
    }
    i2c_Init( (unsigned int*)(((unsigned char *)(g4->pci_io))+I2C0_BASE) );
    init_320aic14k((unsigned int*)(((unsigned char *)(g4->pci_io))+I2C0_BASE)); 

    i2c_Init((unsigned int*)(((unsigned char *)(g4->pci_io))+I2C1_BASE));
    init_320aic14k((unsigned int*)(((unsigned char *)(g4->pci_io))+I2C1_BASE)); 

    i2c_Init((unsigned int*)(((unsigned char *)(g4->pci_io))+I2C2_BASE));
    init_320aic14k((unsigned int*)(((unsigned char *)(g4->pci_io))+I2C2_BASE)); 

    i2c_Init((unsigned int*)(((unsigned char *)(g4->pci_io))+I2C3_BASE));
    init_320aic14k((unsigned int*)(((unsigned char *)(g4->pci_io))+I2C3_BASE)); 
        
    return 0;
}

static void __devexit g4_remove(struct pci_dev *pdev)
{
    struct g4_card *g4 = pci_get_drvdata(pdev);
    struct g4_span *myspan = NULL;
    unsigned long flags;
    unsigned long pci_io_phys;
    unsigned long iomem_size;
    void *pci_io;
    int i = 0;
    
    if (g4) {
        
        __g4_outl__(g4, SWAP_MEM_BASE, 0, 0);                /* disable generate interrupts; */
        __g4_outl__(g4, IRQ_BASE, 0x0050/sizeof(int), 0);    /* disable interrupt enable reg; */

        g4_switch_off_all(g4, 1);
        
        spin_lock_irqsave(&g4->lock,flags);
        g4->dead = 1;
        
        pci_io = g4->pci_io;
        pci_io_phys = g4->pci_io_phys;
        iomem_size = g4->iomem_size;
        g4->pci_io = 0;
        spin_unlock_irqrestore(&g4->lock,flags);

        for (i = 0; i < g4->g4spans; i ++) {
            myspan = &g4->spans[i];
            zt_unregister(&myspan->span);
            if (debug & DEBUG_G4_MSG) {
                printk(KERN_DEBUG "opvxg4xx: unregistered card %d span %d.\n",g4->cardno,i);
            }
        }

        iounmap((void *) pci_io);

        /* release mem region */
        release_mem_region(pci_io_phys, iomem_size);

        spin_lock_irqsave(&g4->lock,flags);
        /* free irq */
        free_irq(g4->irq,g4);
        /* init PCI_COMMAND */
        pci_write_config_word(pdev, PCI_COMMAND, 0);  
        /* disable pci device */
        pci_disable_device(pdev);
        spin_unlock_irqrestore(&g4->lock,flags);

        /* kfree g4 */
        kfree(g4);
        g4 = NULL;
        
        printk(KERN_INFO "opvxg4xx: shutdown OpenVox G4XX GSM/CDMA cards.\n");
    }
}

static struct pci_device_id g4_ids[] = {
    { PCI_VENDOR_ID_OPENVOX, PCI_DEVICE_ID_OPENVOX, PCI_VENDOR_ID_OPENVOX, PCI_SUBSYSTEM_ID_OPENVOX, 0, 0, 0 },
    { 0, }
};

static struct pci_driver g4_driver = {
    .name = "opvxg4xx",
    .probe = g4_probe,
    .remove = __devexit_p(g4_remove),
    .id_table = g4_ids,
};

static int __init g4_init(void)
{
    if (zap_pci_module(&g4_driver))
        return -ENODEV;

    return 0;
}

static void __exit g4_cleanup(void)
{
    pci_unregister_driver(&g4_driver);
}

#ifdef LINUX26
module_param(debug, int, 0600);
module_param(baudrate, long, 0600);
module_param(adcgain, int, 0600);           /* adc gain of codec, in db, from -42 to 20, mute=21; */
module_param(dacgain, int, 0600);           /* dac gain of codec, in db, from -42 to 20, mute=21; */
module_param(sidetone, int, 0600);          /* digital side tone, 0=mute, others can be -3 to -21db, step by -3; */
module_param(inputgain, int, 0600);         /* input buffer cain, can be 0, 6, 12,24db. */
module_param(modem, charp, 0600);           /* modem */
module_param(burn, int, 0600);          /* reset flag */
#else
MODULE_PARM(debug,"i");
MODULE_PARM(baudrate,"l");
MODULE_PARM(adcgain,"i");       /* adc gain of codec, in db, from -42 to 20, mute=21; */
MODULE_PARM(dacgain,"i");       /* dac gain of codec, in db, from -42 to 20, mute=21; */
MODULE_PARM(sidetone,"i");      /* digital side tone, 0=mute, others can be -3 to -21db, step by -3; */
MODULE_PARM(inputgain,"i");     /* input buffer cain, can be 0, 6, 12,24db. */
MODULE_PARM(modem,"s");     /* input buffer cain, can be 0, 6, 12,24db. */
MODULE_PARM(burn,"i");      /* reset flag */
#endif

MODULE_DESCRIPTION("OpenVox GSM/CDMA card zaptel driver");
MODULE_AUTHOR("Miao Lin <lin.miao@openvox.cn>");
#if defined(MODULE_ALIAS)
MODULE_ALIAS("opvxg4xx");
#endif
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

MODULE_DEVICE_TABLE(pci, g4_ids);
module_init(g4_init);
module_exit(g4_cleanup);
